from typing import Union

from colour import Color


class Segment:
    def __init__(self, len: int) -> None:
        """
        Initializes a Segment object with a given length.

        Args:
            len (int): The length of the segment.

        Returns:
            None
        """
        self._colors = [Color("black") for _ in range(len)]
        self._prev_colors = [None for _ in range(len)]
        pass

    def _set_pixel(self, pixel: int, color: Color) -> None:
        """
        Sets the color of a pixel at a given index.

        Args:
            pixel (int): The index of the pixel.
            color (Color): The color to set.

        Returns:
            None

        Raises:
            IndexError: If the pixel index is out of range.
        """
        if pixel < 0 or pixel >= len(self):
            raise IndexError(f"Pixel index {pixel} out of range")
        self._colors[pixel] = color

    def __setitem__(self, key: int, value: Color) -> None:
        """
        Sets the color of a pixel using the [] operator.

        Args:
            key (int): The index of the pixel.
            value (Color): The color to set.

        Returns:
            None
        """
        self._set_pixel(key, value)

    def __getitem__(self, key: int) -> Color:
        """
        Gets the color of a pixel using the [] operator.

        Args:
            key (int): The index of the pixel.

        Returns:
            Color: The color of the pixel.
        """
        return self._colors[key]

    def __delitem__(self, key: int) -> None:
        """
        Raises an error when trying to delete a pixel from the segment.

        Args:
            key (int): The index of the pixel.

        Returns:
            None

        Raises:
            NotImplementedError: Deleting pixels from a segment is not supported.
        """
        raise NotImplementedError("Cannot delete pixels from a segment")

    def __len__(self) -> int:
        """
        Returns the length of the segment.

        Returns:
            int: The length of the segment.
        """
        return self._colors.__len__()

    def set_all(self, color: Color) -> None:
        """
        Sets all pixels in the segment to a given color.

        Args:
            color (Color): The color to set.

        Returns:
            None
        """
        for i in range(len(self)):
            self._set_pixel(i, color)

    def set_range(self, start: int, stop: int, color: Color) -> None:
        """
        Sets a range of pixels in the segment to a given color.

        Args:
            start (int): The starting index of the range.
            stop (int): The ending index of the range.
            color (Color): The color to set.

        Returns:
            None

        Raises:
            IndexError: If the start or stop index is out of range.
        """
        if start < 0 or start >= len(self) or stop < 0 or stop >= len(self):
            raise IndexError(f"Pixel index {start} or {stop} out of range")
        for i in range(start, stop + 1):
            self._set_pixel(i, color)

    def set_off(self) -> None:
        """
        Sets all pixels in the segment to black (off).

        Returns:
            None
        """
        self.set_all(Color("black"))

    def _get_changes(self) -> list[Color]:
        """
        Gets the changes in colors between the current state and the previous state.

        Returns:
            list[Color]: A list of color changes.
        """
        changes = [None] * len(self)
        for i in range(len(self)):
            changes[i] = (
                self._colors[i] if self._colors[i] != self._prev_colors[i] else None
            )

        return changes

    def reset_changes(self) -> None:
        """
        Resets the changes in colors.

        Returns:
            None
        """
        self._prev_colors = self._colors.copy()

    def get_optimised_colour_changes(self) -> Union[int, Color]:
        """
        Returns a list with the optimised colour changes according to
        https://kno.wled.ge/interfaces/json-api/#per-segment-individual-led-control.

        Returns:
            Union[int, Color]: A list with the optimised colour changes.
        """
        changes = [(c, 1) for c in self._get_changes()]
        complete_sequence: Union[int, Color] = []

        last_unique_element = 0
        # group identical colours
        for i in range(1, len(changes)):
            if changes[i][0] and changes[i][0] == changes[last_unique_element][0]:
                changes[last_unique_element] = (
                    changes[i][0],
                    changes[last_unique_element][1] + 1,
                )
                changes[i] = (None, 0)
            else:
                last_unique_element = i

        sequence_started = False
        for i in range(len(changes)):
            if changes[i][1] == 0:
                continue

            # empty spaces, break the sequence
            if not changes[i][0]:
                sequence_started = False
                continue

            # a valid colour, check repetitions
            stop_index = i + changes[i][1] - 1
            colour = changes[i][0]
            # if single element, check if start index was already added
            if stop_index == i:
                if not sequence_started:
                    sequence_started = True
                    complete_sequence.append(i)
            else:
                # range: add start and stop index
                complete_sequence.extend([i, stop_index])
                sequence_started = False
            complete_sequence.append(colour)

        return complete_sequence



from dataclasses import dataclass
from typing import List, Optional, Union

from dataclasses_json import DataClassJsonMixin, config, dataclass_json


class JsonSerializableMixin(DataClassJsonMixin):
    """
    Mixin class for JSON serialization and deserialization of data classes.
    """

    dataclass_json_config = config(exclude=lambda f: f is None)["dataclasses_json"]


@dataclass
class Leds(JsonSerializableMixin):
    """
    Represents LED properties.
    """

    count: Optional[int] = None
    rgbw: Optional[bool] = None
    pin: Optional[List[int]] = None
    pwr: Optional[int] = None
    maxpwr: Optional[int] = None
    maxseg: Optional[int] = None


@dataclass_json
@dataclass
class Info(JsonSerializableMixin):
    """
    Represents information about the WLED device.
    """

    ver: Optional[str] = None
    vid: Optional[int] = None
    leds: Optional[Leds] = None
    name: Optional[str] = None
    udpport: Optional[int] = None
    live: Optional[bool] = None
    fxcount: Optional[int] = None
    palcount: Optional[int] = None
    arch: Optional[str] = None
    core: Optional[str] = None
    freeheap: Optional[int] = None
    uptime: Optional[int] = None
    opt: Optional[int] = None
    brand: Optional[str] = None
    product: Optional[str] = None
    btype: Optional[str] = None
    mac: Optional[str] = None


@dataclass
class Nl(JsonSerializableMixin):
    """
    Represents Nightlight properties.
    """

    on: Optional[bool] = None
    dur: Optional[int] = None
    mode: Optional[int] = None
    tbri: Optional[int] = None
    rem: Optional[int] = None


@dataclass
class Udpn(JsonSerializableMixin):
    """
    Represents UDP notifications properties.
    """

    send: Optional[bool] = None
    recv: Optional[bool] = None
    sgrp: Optional[int] = None
    rgrp: Optional[int] = None


@dataclass
class Seg(JsonSerializableMixin):
    """
    Represents a segment of LEDs.
    """

    id: Optional[int] = None
    start: Optional[int] = None
    stop: Optional[int] = None
    len: Optional[int] = None
    grp: Optional[int] = None
    spc: Optional[int] = None
    of: Optional[int] = None
    on: Optional[bool] = None
    frz: Optional[bool] = None
    bri: Optional[int] = None
    cct: Optional[int] = None
    set: Optional[int] = None
    col: Optional[List[List[int]]] = None
    fx: Optional[int] = None
    sx: Optional[int] = None
    ix: Optional[int] = None
    pal: Optional[int] = None
    c1: Optional[int] = None
    c2: Optional[int] = None
    c3: Optional[int] = None
    sel: Optional[bool] = None
    rev: Optional[bool] = None
    mi: Optional[bool] = None
    o1: Optional[bool] = None
    o2: Optional[bool] = None
    o3: Optional[bool] = None
    si: Optional[int] = None
    m12: Optional[int] = None
    i: Optional[List[Union[int, str]]] = None


@dataclass
class State(JsonSerializableMixin):
    """
    Represents the state of the WLED device.
    """

    off: Optional[bool] = None
    bri: Optional[int] = None
    transition: Optional[int] = None
    ps: Optional[int] = None
    pl: Optional[int] = None
    nl: Optional[Nl] = None
    udpn: Optional[Udpn] = None
    lor: Optional[int] = None
    mainseg: Optional[int] = None
    seg: Optional[List[Seg]] = None


@dataclass
class WledData(JsonSerializableMixin):
    """
    Represents the data structure for WLED device.
    """

    info: Optional[Info] = None
    effects: Optional[List[str]] = None
    palettes: Optional[List[str]] = None
    state: Optional[State] = None



from abc import ABC, abstractmethod
from time import sleep
from typing import override

from loguru import logger
from serial import Serial
from serial.threaded import LineReader, Protocol, ReaderThread
from serial.tools import list_ports


class Connection(ABC):
    """Abstract base class for a connection."""

    def __init__(self):
        pass

    @abstractmethod
    def send(self, data: str) -> str:
        """Send data over the connection."""
        pass

    @abstractmethod
    def receive(self) -> str:
        """Receive data from the connection."""
        pass

    @abstractmethod
    def get_wled_status(self) -> str:
        """Get the status of the connected device."""
        pass

    def start(self):
        """Start the connection."""
        pass

    def stop(self):
        """Stop the connection."""
        pass


class SerialReader(LineReader):
    """Serial reader class for reading lines from a serial connection."""

    def __init__(self):
        super(SerialReader, self).__init__()
        self.received_data: list[str] = []

    def connection_made(self, transport):
        super(SerialReader, self).connection_made(transport)
        logger.info("port opened")

    def handle_line(self, data):
        logger.trace("line received: {}\n".format(repr(data)))
        self.received_data.append(data)

    def connection_lost(self, exc):
        if exc:
            logger.error(exc)
        logger.info("port closed")


class SerialConnection(Connection):
    """Serial connection class for communicating with a device over a serial port."""

    TIMEOUT = 0.05
    COMMAND_WAIT = 0.005

    @override
    def __init__(self, port: str, baudrate: int):
        super(SerialConnection, self).__init__()
        self._port = port
        self._baudrate = baudrate
        self._serial: Serial = None
        self._protocol: Protocol = None
        self.start()

    @override
    def send(self, data: str) -> None:
        """Send data over the serial connection."""
        self._clear_buffer()
        self._send_string(data)
        sleep(self.COMMAND_WAIT)

    @override
    def receive(self) -> str:
        """Receive data from the serial connection."""
        return self._retrieve_response()

    def _send_string(self, data: str) -> None:
        """Send a string over the serial connection."""
        self._protocol.write_line(data)

    def _retrieve_response(self):
        """Retrieve the response from the serial connection."""
        timeout = 20
        while not self._protocol.received_data and timeout > 0:
            timeout -= 1
            sleep(self.COMMAND_WAIT)
        ret = self._protocol.received_data.pop()
        return ret

    def _clear_buffer(self):
        """Clear the buffer of received data."""
        self._protocol.received_data.clear()

    @classmethod
    def available_ports(cls, regexp: str = None, include_links=True) -> list[any]:
        """Get a list of available serial ports."""
        return list_ports.grep(regexp, include_links)

    @override
    def get_wled_status(self) -> str:
        """Get the status of the serial connection."""
        status_request = '{"v":true}'
        self._clear_buffer()
        self._send_string(status_request)
        return self._retrieve_response()

    @override
    def start(self):
        """Start the serial connection."""
        self._serial = Serial(
            self._port, baudrate=self._baudrate, timeout=SerialConnection.TIMEOUT
        )
        self._protocol = ReaderThread(self._serial, SerialReader).__enter__()

    @override
    def stop(self):
        """Stop the serial connection."""
        self._protocol.__exit__(None, None, None)



from colour import Color

from wled_adapter.connection import Connection
from wled_adapter.segment import Segment
from wled_adapter.state import Seg, State, WledData


class Adapter:
    """
    Represents an adapter for controlling WLED segments using a specified interface.

    Args:
        interface (Connection): The interface used for communication with the WLED
        device.

    Attributes:
        _interface (Connection): The interface used for communication with the WLED
        device.
        _segments (dict[int, Segment]): A dictionary of segments, where the key is the
        segment ID and the value is a Segment object.

    """

    def __init__(self, interface: Connection):
        self._interface = interface
        self._segments: dict[int, Segment] = {}

    def init(self):
        """
        Initializes the adapter by starting the interface.
        """
        self._interface.start()

    def cleanup(self):
        """
        Cleans up the adapter by stopping the interface.
        """
        self._interface.stop()

    @property
    def segments(self):
        """
        Returns the dictionary of segments.
        """
        return self._segments

    def initialise_segments(self):
        """
        Initializes the segments based on the current status received from the WLED
        device.
        """
        state = self.get_state()
        # reset the current segment state
        self._segments = {}

        for segment in state.state.seg:
            segment_len = 0
            if hasattr(segment, "len"):
                segment_len = segment.len
            else:
                segment_len = segment.stop - segment.start
            self._segments[segment.id] = Segment(segment_len)

    def get_state(self) -> WledData:
        """
        Retrieves the state of the WLED device.

        Returns:
            WledData: The state of the WLED device.
        """
        status_json = self._interface.get_wled_status()
        status = WledData.from_json(status_json)
        return status

    def update_segment(self, segment_id: int):
        """
        Updates the specified segment colour state on the WLED device.

        Args:
            segment_id (int): The ID of the segment to update.
        """
        segment = self._segments[segment_id]
        colors = segment.get_optimised_colour_changes()

        colors = [Color(c).hex_l[1:] if isinstance(c, Color) else c for c in colors]
        request = State(seg=[Seg(id=segment_id, i=colors)])
        self.send_state(request)
        segment.reset_changes()

    def send_state(self, state: State):
        """
        Sends the specified state to the WLED device.

        Args:
            state (State): The state to send.
        """
        json = state.to_json(skipkeys=True)
        self._interface.send(json)
